/************************************************************************
*  Hansen.C (Test example for linear lists)
*
*  Version:                     1.4
*  Date:                        2/1/94
*  Author:                      O. Knueppel
*************************************************************************/

#include "util.h"
#include "func.h"

// Define a list where each list element contains
// an interval and an interval vector.
// The list element is called BOXPAIR and the
// list type is called LIST.

class BOXPAIR {
  INTERVAL range;
  INTERVAL_VECTOR box;
public:
  BOXPAIR () {}
  BOXPAIR (INTERVAL & ra, INTERVAL_VECTOR & bo)
    { range = ra; box = bo; }
  BOXPAIR & operator = (BOXPAIR & v)
    { range = v.range; box = v.box; return *this; }
  friend INTERVAL & Range (BOXPAIR & v)
    { return v.range; }
  friend INTERVAL_VECTOR & Box (BOXPAIR & v)
    { return v.box; }
  friend ostream & operator << (ostream & o, BOXPAIR & v)
    { return o << v.box << "->" << v.range; }
};

#define LIST_ELEMENT BOXPAIR
#define LIST         SOLUTIONLIST
#define LISTOBJECT   SOLUTIONLISTOBJECT
#include "linlist.hge"
#include "linlist.cge"
#undef LIST_ELEMENT
#undef LIST
#undef LISTOBJECT

// Some functions which are needed for this example

VOID Cleanup (SOLUTIONLIST & list, REAL threshold)
// delete all entries in "list" with Inf(Range (y)) > threshold.
{
  if (IsEmpty (list)) return;
  First (list);
  while (!Finished (list))
    {
      if (Inf( Range (Current (list))) > threshold) RemoveCurrent (list);
      else Next (list);
    }
}

INT MaxDiam (INTERVAL_VECTOR & v)
// Return the direction with the largest diameter
{
  INT i, k = 1;
  REAL md = Diam (v(1));
  REAL d;

  for (i = 2; i <= Dimension (v); i++)
    if ((d = Diam(v(i))) > md) { k = i; md = d; }
  return k;
}

VOID HansenMidpointTest (INTERVAL_VECTOR & x, INTERVAL (*pfunc)(INTERVAL_VECTOR &),
			 SOLUTIONLIST & list, REAL eps)
// Find the global minimum of the function given by "pfunc"
// over the box "x" and return a list of subboxes, where
// the relative diameter of each subbox is less than "eps"
{
  INT k;
  REAL fs, fs2;
  BOXPAIR y1, y2;
  BOXPAIR y((*pfunc) (x), x);

  fs = Sup ( (*pfunc) (Hull (Mid (x))));
  list += y;
  while (Max (RelDiam (Box (y))) > eps)
    {
      k = MaxDiam (Box (y));
      Box (y1) = Lower (Box (y), k);
      Box (y2) = Upper (Box (y), k);
      Range (y1) = (*pfunc) (Box (y1));
      Range (y2) = (*pfunc) (Box (y2));
      list--;
      list += y1; list += y2;
      if (!IsEmpty (list))
	{
	  y = First (list);
	  fs2 = Sup ((*pfunc) (Hull (Mid (Box (y)))));
	  if (fs2 < fs)
	    {
	      fs = fs2;
	      Cleanup (list, fs);
	    }
	}
    }
}

INTERVAL f (INTERVAL_VECTOR& x)
// The function to be minimized
{
  return 0.5 * Sqr(Sqr(x(1)) - x(2)) + Sqr(x(1) - 1.0);
}

INT main()
{
  SOLUTIONLIST list;
  INTERVAL_VECTOR v(2);

  Initialize (v, Hull (-2.0, 8.0)); // initial box

  cout << "SIMPLE GLOBAL MINIMZATION ALGORITHM" << endl;
  cout << "The output should be a list with elements of the kind" << endl;
  cout << "  interval vector -> interval" << endl;
  cout << "The convex hull of all interval vectors contains all global" << endl;
  cout << "minimum points of f in " << v << endl;
  cout << "The convex hull of all intervals contains an enclosure of the" << endl;
  cout << "global minimum value of f." << endl;
  cout << "f = 0.5 * (x^2 - y)^2 + (x - 1)^2" << endl;
  cout << "global minimum point: (1,1); global minimum value: 0" << endl << endl;

  HansenMidpointTest (v, f, list, 1e-6);

  cout.precision(16);
  cout << "Results:" << endl;
  cout << list << endl;
  return 0;
}
